; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Fonts managing library.
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


proc FontCreate, .fontname, .size, .weight, .flags
begin
        push    ebx ecx edx
        stdcall StrDup, [.fontname]
        mov     ebx, eax

        test    [.flags], ffMonospaced
        jz      @f
        stdcall StrCat, ebx, FontAttr.mono
@@:
        sub     [.size], 2
        cmp     [.size], 0
        jle     @f

        stdcall StrCat, ebx, FontAttr.size
        stdcall NumToStr, [.size], ntsDec or ntsUnsigned
        push    eax
        stdcall StrCat, ebx, eax
        stdcall StrDel ; from the stack
@@:
        cmp     [.weight], 0
        je      @f
        shr     [.weight], 2
        stdcall StrCat, ebx, FontAttr.weight
        stdcall NumToStr, [.weight], ntsDec or ntsUnsigned
        push    eax
        stdcall StrCat, ebx, eax
        stdcall StrDel ; from the stack

@@:
        test    [.flags], ffItalic
        jz      @f
        stdcall StrCat, ebx, FontAttr.italic
@@:

        stdcall StrPtr, ebx
        cinvoke XftFontOpenName, [hApplicationDisplay], 0, eax

        stdcall StrDel, ebx
        pop     edx ecx ebx
        return
endp


iglobal
  if used FontCreate
    label FontAttr byte
    .mono db ',Mono', 0
    .size db ':pixelsize=', 0
    .weight db ':weight=', 0
    .italic db ':slant=italic', 0
  end if
endg










macro XftPushArg field, type, value {
      push value
      push type
      push field
      add  ebx, 12
}

macro XftStartArguments {
      push 0
      mov  ebx, 4
}

XftFamily  text 'family'
XftStyle   text 'style'
XftSpacing text 'spacing'
XftSlant   text 'slant'
XftWeight  text 'weight'
XftWidth   text 'width'
XftPixelSize text 'pixelsize'
XftAntialias text 'antialias'
XftAutohint  text 'autohint'
XftHintStyle text 'hintstyle'
XftHinting   text 'hinting'
XftRGBA      text 'rgba'


proc FontCreate3, .fontname, .size, .weight, .flags
begin
        push    ebx ecx edx


        XftStartArguments

        cmp     [.fontname], 0
        je      @f
        stdcall StrPtr, [.fontname]
        XftPushArg XftFamily, XftTypeString, eax
@@:

        cmp     [.size], 2
        jle     @f
        sub     [.size], 2
        XftPushArg XftPixelSize, XftTypeInteger, [.size]
@@:

        cmp     [.weight], 0
        jle     @f
        shr     [.weight], 1
        XftPushArg XftWeight,    XftTypeInteger, [.weight]
@@:

        test    [.flags], ffItalic
        jz      @f
        XftPushArg XftSlant, XftTypeInteger, XFT_SLANT_ITALIC
@@:

        test    [.flags], ffMonospaced
        jz      @f
        XftPushArg XftSpacing, XftTypeInteger, XFT_MONO
@@:
        XftPushArg XftAntialias, XftTypeBool, 0
        XftPushArg XftHinting, XftTypeBool, 0
        XftPushArg XftHintStyle, XftTypeInteger, 0
        XftPushArg XftRGBA, XftTypeInteger, 5
        XftPushArg XftWidth, XftTypeInteger, 100

        invoke  XftFontOpen, [hApplicationDisplay], 0
        add     ebx, 8
        add     esp, ebx

        pop     edx ecx ebx
        return
endp




proc FontDestroy, .font
begin
        stdcall XftFontClose, [hApplicationDisplay], [.font]
        return
endp