; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TProgressBar object class; Simple progress bar control.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________
module "TProgressbar library"

ObjectClass Progress,                   \
            Window,                     \
            0,                          \
            0,                          \
            TProgress.Get,              \
            TProgress.Set,              \
            TProgress.ExecCmd,          \
            TProgress.SysEventHandler

object TProgress, TWindow
  ._Pos   dd ?
  ._Max   dd ?
  ._Color dd ?

  param .Pos
  param .Max
  param .Color

  method .Step
endobj



proc TProgress.Get, .pobj, .paramID
begin
        push    ecx

        stdcall IsObject, [.pobj], CProgress
        jc      .exit

        mov     eax, [.pobj]
        mov     ecx, [.paramID]

        cmp     ecx, TProgress.Pos
        jb      .exit
        cmp     ecx, TProgress.Color
        ja      .exit

        sub     ecx, TProgress.Pos
        pushd   [eax+TProgress._Pos+4*ecx]
        pop     eax

        clc
        pop     ecx
        return

.exit:
        stc
        pop     ecx
        return
endp



proc TProgress.Set, .pobj, .paramID, .value
begin
        push    eax ecx

        stdcall IsObject, [.pobj], CProgress
        jc      .exit

        mov     eax, [.pobj]
        mov     ecx, [.paramID]

        cmp     ecx, TProgress.Pos
        jb      .exit
        cmp     ecx, TProgress.Color
        ja      .exit

        sub     ecx, TProgress.Pos
        pushd   [.value]
        popd    [eax+TProgress._Pos+4*ecx]

        execute [.pobj], TProgress.Refresh

        clc
        pop     ecx eax
        return

.exit:
        stc
        pop     ecx eax
        return
endp


proc TProgress.SysEventHandler, .pobj, .pEvent
begin
        push    eax ebx ecx esi

        mov     ebx, [.pEvent]
        cmp     [ebx+TSysEvent.event], sePaint
        je      .paint

.finish:
        stc
        pop     esi ecx ebx eax
        return

.paint:
locals
  .bounds TBounds
  .style TLineStyle
endl
        mov     [.bounds.x], 0
        mov     [.bounds.y], 0
        mov     eax, [esi+TProgress._width]
        mov     ecx, [esi+TProgress._height]
        mov     [.bounds.width], eax
        mov     [.bounds.height], ecx

        lea     eax, [.bounds]
        stdcall [DrawBox], [ebx+TPaintEvent.context], eax, 0, bxSunken or bxNoFill

        stdcall DrawRectangle, [ebx+TPaintEvent.context], eax, [clDialogBk]

        inc     [.bounds.x]
        inc     [.bounds.y]
        sub     [.bounds.width], 2
        sub     [.bounds.height], 2

        mov     eax, [esi+TProgress._Pos]
        mov     ecx, [esi+TProgress._Max]
        test    ecx, ecx
        jnz     @f
        inc     ecx
@@:
        cmp     eax, ecx
        jle     @f
        mov     eax, ecx
@@:
        imul    eax, [.bounds.width]
        cdq
        idiv    ecx

        stdcall DrawFillRect, [ebx+TPaintEvent.context], [.bounds.x], [.bounds.y], eax, [.bounds.height], [esi+TProgress._Color]
        sub     [.bounds.width], eax
        inc     eax
        inc     [.bounds.width]
        stdcall DrawFillRect, [ebx+TPaintEvent.context], eax, [.bounds.y], [.bounds.width], [.bounds.height], [clDialogBk]

        jmp     .finish
endp




proc TProgress.ExecCmd, .self, .method
begin
        cmp     [.method], TProgress.Step
        je      .step
        stc
        return

;-------------------------------------------------------------------------
.step:
        push    eax ecx

        mov     eax, [.self]
        mov     ecx, [eax+TProgress._Pos]
        cmp     ecx, [eax+TProgress._Max]
        jge     .exit

        inc     [eax+TProgress._Pos]
        execute eax, TProgress.Refresh

.exit:
        pop     ecx eax
        clc
        return
endp



endmodule
