; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Window management OS interface functions.
;
;  Target OS: Win32
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________

iglobal
if used cWindowClassName
  cWindowClassName du 'FreshWin', 0
end if
endg

nWindowExtraBytes = 32
ofsWindowStruct = nWindowExtraBytes - 4

;_________________________________________________________________________________________
;
; Registers common window class
; Call only once
;_________________________________________________________________________________________

if used _CreateNullWindow
initialize __RegisterWindowClass

.wc WNDCLASS

begin
        xor     eax, eax
        lea     edi, [.wc]
        mov     ecx, sizeof.WNDCLASS / 4
        rep stosd

        mov     [.wc.style], CS_OWNDC or CS_DBLCLKS; or CS_VREDRAW or CS_HREDRAW
        mov     [.wc.lpfnWndProc], __CommonWindowProc
        mov     [.wc.cbWndExtra], nWindowExtraBytes

;        invoke  CreateSolidBrush, $0000ff
;        mov     [.wc.hbrBackground], eax

        mov     eax,[hInstance]
        mov     [.wc.hInstance],eax
        mov     [.wc.lpszClassName], cWindowClassName

        lea     eax, [.wc]
        invoke  RegisterClassW, eax

        return
endp
end if



;_________________________________________________________________________________________


proc _CreateNullWindow
begin
        push    ecx edx
        xor     ecx, ecx
        invoke  CreateWindowExW, ecx, cWindowClassName, ecx, WS_CLIPSIBLINGS or WS_CLIPCHILDREN, ecx, ecx, ecx, ecx, ecx, ecx, ecx, ecx
        pop     edx ecx
        return
endp

;_________________________________________________________________________________________



proc _DestroyWindow, .hwnd
begin
        push    eax ecx edx
        invoke  DestroyWindow, [.hwnd]
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _GetParent, .hwnd
begin
        push    ecx edx
;        invoke  GetParent, [.hwnd]
        invoke  GetWindowLongW, [.hwnd], GWL_HWNDPARENT
        pop     edx ecx
        return
endp

;_________________________________________________________________________________________


proc _GetChildren, .hwnd
begin
        push    ebx ecx edx

        xor     ebx, ebx

        invoke  GetWindow, [.hwnd], GW_CHILD
        mov     ecx, eax
        jecxz   .endchildren

        stdcall CreateArray, sizeof.TWinArrayItem
        mov     ebx, eax

.loophwnd:
        stdcall AddArrayItems, ebx, 1
        mov     ebx, edx

        mov     [eax+TWinArrayItem.handle], ecx

        invoke  GetWindow, ecx, GW_HWNDNEXT
        mov     ecx, eax
        jecxz   .endchildren
        jmp     .loophwnd

.endchildren:
        mov     eax, ebx
        pop     edx ecx ebx
        return
endp


;_________________________________________________________________________________________



proc _GetVisible, .hwnd
begin
        push    ecx edx
        invoke  IsWindowVisible, [.hwnd]
        pop     edx ecx
        return
endp


;_________________________________________________________________________________________



proc _GetWindowBounds, .hwnd, .pBounds
begin
        push    eax ecx edx

        invoke  GetClientRect, [.hwnd], [.pBounds]
        invoke  GetParent, [.hwnd]
        invoke  MapWindowPoints, [.hwnd], eax, [.pBounds], 2

        mov     edx, [.pBounds]
        mov     eax, [edx+TBounds.width]
        mov     ecx, [edx+TBounds.height]
        sub     eax, [edx+TBounds.x]
        sub     ecx, [edx+TBounds.y]
        mov     [edx+TBounds.width], eax
        mov     [edx+TBounds.height], ecx

        pop     edx ecx eax
        return
endp



;_________________________________________________________________________________________


proc _SetWindowBounds, .hwnd, .pBounds
.rect RECT
begin
        push    eax ecx edx

        mov     eax, [.pBounds]
        push    [eax+TBounds.width] [eax+TBounds.height]
        pop     [.rect.bottom] [.rect.right]
        mov     [.rect.left], 0
        mov     [.rect.top], 0

        invoke  GetWindowLongW, [.hwnd], GWL_EXSTYLE
        push    eax
        push    FALSE
        invoke  GetWindowLongW, [.hwnd], GWL_STYLE
        push    eax

        lea     eax, [.rect]
        invoke  AdjustWindowRectEx, eax ; and from the stack

        mov     eax, [.rect.right]
        mov     ecx, [.rect.bottom]
        sub     eax, [.rect.left]
        sub     ecx, [.rect.top]

        mov     edx, [.pBounds]
        invoke  SetWindowPos, [.hwnd], 0, [edx+TBounds.x], [edx+TBounds.y], eax, ecx, SWP_NOZORDER or SWP_NOREDRAW

        pop     edx ecx eax
        return
endp





;_________________________________________________________________________________________
wsNoneWindowBorder = WS_POPUP
wsFullWindowBorder = WS_BORDER or WS_CAPTION or WS_SYSMENU or WS_SIZEBOX or WS_MINIMIZEBOX or WS_MAXIMIZEBOX
wsModalWindowBorder  = WS_CAPTION or WS_SYSMENU
wsToolboxWindowBorder = WS_BORDER or WS_CAPTION or WS_SYSMENU or WS_SIZEBOX

wsAllWindowBorder = wsNoneWindowBorder or wsFullWindowBorder or wsModalWindowBorder or wsToolboxWindowBorder

proc _SetWindowBorder, .hwnd, .brdType
.bounds TBounds
begin
        push    eax ebx ecx edx esi

        invoke  GetWindowLongW, [.hwnd], GWL_STYLE
        mov     ebx, eax

        mov     esi, [.brdType]
        and     esi, 3

        mov     ecx, [.styles+4*esi]
        mov     edx, [.exstyles+4*esi]

        and     ebx, not wsAllWindowBorder
        test    ebx, WS_CHILD
        jz      @f
        and     ecx, not (WS_CAPTION or WS_SYSMENU or WS_SIZEBOX)
        and     edx, not (WS_EX_APPWINDOW or WS_EX_DLGMODALFRAME or WS_EX_TOOLWINDOW)
@@:
        or      ebx, ecx
        invoke  SetWindowLongW, [.hwnd], GWL_EXSTYLE, edx
        invoke  SetWindowLongW, [.hwnd], GWL_STYLE, ebx

        lea     esi, [.bounds]
        stdcall _GetWindowBounds, [.hwnd], esi
        stdcall _SetWindowBounds, [.hwnd], esi

        pop     esi edx ecx ebx eax
        return

.styles dd wsNoneWindowBorder
        dd wsFullWindowBorder
        dd wsModalWindowBorder
        dd wsToolboxWindowBorder

.exstyles dd 0
          dd WS_EX_APPWINDOW
          dd WS_EX_DLGMODALFRAME
          dd WS_EX_TOOLWINDOW

endp

;_________________________________________________________________________________________

proc _ShowWindow, .hwnd, .flag
begin
        push    eax ecx edx
        invoke  ShowWindow, [.hwnd], [.flag]
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _RefreshWindow, .hwnd
begin
        push    eax ecx edx
        xor     eax, eax
;        invoke  InvalidateRect, [.hwnd], eax, eax
        invoke  RedrawWindow, [.hwnd], eax, eax, RDW_INVALIDATE or RDW_UPDATENOW or RDW_ERASENOW

        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _SetFocus, .hwnd
begin
        push    eax ecx edx
        invoke  SetFocus, [.hwnd]
        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________

proc _AddChild, .hwnd, .child
begin
        push    eax ecx edx

        invoke  GetWindowLongW, [.child], GWL_STYLE
        or      eax, WS_CHILD
        and     eax, not (WS_DLGFRAME or WS_BORDER)
        invoke  SetWindowLongW, [.child], GWL_STYLE, eax
        invoke  SetParent, [.child], [.hwnd]

        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________


; Common utility procedures.


;_________________________________________________________________________________________
;
; Returns the window TObject structure, from the window handle.
;_________________________________________________________________________________________

proc _GetWindowStruct, .hwin
begin
        push    ecx edx
        invoke  GetWindowLongW, [.hwin], ofsWindowStruct
        pop     edx ecx
        return
endp



;_________________________________________________________________________________________



proc _SetWindowStruct, .hwin, .value
begin
        push    eax ecx edx
        invoke  SetWindowLongW, [.hwin], ofsWindowStruct, [.value]
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________



proc _SetWindowTextUtf8, .hwnd, .ptrUtf8
.pwc dd ?
.charlen dd ?
begin
        stdcall StrLen, [.ptrUtf8]
        lea     ecx, [eax*4]

        stdcall GetMem, ecx
        mov     [.pwc], eax

        invoke  MultiByteToWideChar, CP_UTF8, 0, [.ptrUtf8], -1, [.pwc], ecx
        mov     [.charlen], eax

        invoke  SendMessageW, [.hwnd], WM_SETTEXT, 0, [.pwc]
        stdcall FreeMem, [.pwc]
        return
endp



;_________________________________________________________________________________________



proc _GetWindowTextUtf8, .hwnd, .ptrUtf8
.pwc dd ?
.charlen dd ?
begin
        push    ebx ecx edx edi
        stdcall StrNew
        mov     ebx, eax

        invoke  SendMessage, [.hwnd], WM_GETTEXTLENGTH, 0, 0
        mov     [.charlen], eax
        lea     ecx, [eax*4]

        stdcall StrSetCapacity, ebx, ecx
        mov     edi, eax

        stdcall GetMem, ecx
        mov     [.pwc], eax

        invoke  SendMessage, [.hwnd], WM_GETTEXT, [.charlen], [.pwc]
        invoke  WideCharToMultiByte, CP_UTF8, 0, [.pwc], -1, edi, [edi+string.capacity], 0, 0
        stdcall FreeMem, [.pwc]
        stdcall StrFixLen, ebx
        mov     eax, ebx
        pop     edi edx ecx ebx
        return
endp



;_________________________________________________________________________________________


proc _EnableWindow, .hwnd, .flag
begin
        push    eax ecx edx
        invoke  EnableWindow, [.hwnd], [.flag]
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _SetModalTowards, .hwnd, .hwndParent
begin
        push    eax ecx edx
        stdcall _SetWindowBorder, [.hwnd], borderModal
        invoke  SetWindowLongW, [.hwnd], GWL_HWNDPARENT, [.hwndParent]

        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _FinalizeModal, .hwnd, .hwndParent
begin
        return
endp


;_________________________________________________________________________________________


proc __CommonWindowProc, .hwnd, .wmsg, .wparam, .lparam
begin
        stdcall __ProcessOneSystemEvent, [.hwnd], [.wmsg], [.wparam], [.lparam]         ; sometimes windows calls winproc directly instead of using postmessage.
        jnc     .finish

;        invoke  DefWindowProcW, [.hwnd], [.wmsg], [.wparam], [.lparam]

.finish:
        return
endp

