; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Macros for create and call of high level procedures.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


macro locals {
  local ..locsize
  locsize equ ..locsize
  virtual at ebp - ..locsize - .__info.commonframe.size
  @@:
}


macro endl {
  rb (4 - ($- @b) and 11b) and 11b
  locsize = $ - @b
  if $-@b > maxsize
    maxsize = $-@b
  end if
  end virtual
  restore locsize
}


macro err@endp { }


macro proc name, [arg] {
  common
    \err@endp

    macro err@endp \{
      end if
      Missing 'endp' before the procedure.
    \}

  if defined options.AlignCode & options.AlignCode
    align 16
  end if

  name:
    .__info.id = 1
    .__info.start = $

    virtual at ebp+8
      if ~ arg eq
        forward
          arg dd ?
          if ~ defined name#arg
            ERROR! the argument `arg MUST begins with dot
          end if
        common
      end if
        .__info.argsize = $-ebp-8
    end virtual

    if ~ used name
      if defined options.ShowSkipped & options.ShowSkipped
        display 1,'Procedure skipped: ',`name, $0d, $0a
      end if
    else

    virtual at ebp - .__info.commonframe.size
    .__info.commonframe.begin:
}


macro begin {
      rb (4 - ($ - .__info.commonframe.begin) and 11b) and 11b
      .__info.commonframe.size = $ - .__info.commonframe.begin
    end virtual

    local ..maxsize
    maxsize equ ..maxsize
    ..maxsize = 0

    if .__info.framesize
      if defined options.FastEnter & options.FastEnter
        push ebp
        mov  ebp, esp
        sub  esp, .__info.framesize
      else
        enter .__info.framesize, 0
      end if
    else
      if .__info.argsize
        push ebp
        mov  ebp, esp
      end if
    end if
}


macro cret {
  if .__info.argsize | .__info.framesize
    if defined options.FastEnter & options.FastEnter
      mov   esp, ebp
      pop   ebp
    else
      leave
    end if
  end if
  retn
}


macro return {
  if .__info.argsize | .__info.framesize
    if defined options.FastEnter & options.FastEnter
      mov   esp, ebp
      pop   ebp
    else
      leave
    end if
  end if

  if .__info.argsize
    retn .__info.argsize
  else
    retn
  end if
}



macro endp {
    .__info.framesize = maxsize + .__info.commonframe.size
    .__info.codesize = $ - .__info.start
  end if
  restore maxsize
  purge err@endp
}


; dispatch procedure

macro dproc name, [arg] {
common

  \err@endp

  macro err@endp \{
    end if
    Missing 'enddp' before this procedure.
  \}


  if ~ used name
    if defined options.ShowSkipped & options.ShowSkipped
      display 1,'Dispatch procedure skipped: ',`name, $0d, $0a
    end if
  else

  name:
    .__info.id = 2
    .__info.start = $

  virtual at ebp+8
    forward
      arg dd ?
      if ~ defined name#arg
        ERROR! the argument `arg MUST begin with dot
      end if
    common
      .__info.argsize = $ - ebp - 8
  end virtual

  virtual at ebp - .__info.commonframe.size
  .__info.commonframe.begin:
}



macro dispatch arg {  ; begins procedure instructions
  common
    local ..casecounter, ..caselist, ..caseaddr
    casecounter equ ..casecounter
    caselist equ ..caselist
    caseaddr equ ..caseaddr

    ..casecounter = 0

    mov  eax, arg

    ..caselist:
    ..caseaddr = $

    rb .__info.casesize
}


macro oncase idname {
common
     local offset

;     if idname >= -128 & idname <= 127
;       store word $f883 at caseaddr             ; cmp eax, byte imm     ; 3 bytes
;       store byte idname at caseaddr + 2
;       caseaddr = caseaddr + 3
;    else
       store byte $3d at caseaddr               ; cmp eax, dword imm    ; 5 bytes
       store dword idname at caseaddr+1
       caseaddr = caseaddr + 5
;     end if

     offset = $ - caseaddr - 2

;; THIS OPTIMIZATION GENERATES "code can not be compiled" error in FASM and need to be fixed...
;     if offset >= -128 & offset <=127
;        store byte $74 at caseaddr              ; je byte offset        ; 2 bytes
;        store byte offset at caseaddr+1
;        caseaddr = caseaddr + 2
;     else
        store word $840f at caseaddr            ; je dword offset       ; 6 bytes
        store dword offset - 4 at caseaddr+2
        caseaddr = caseaddr + 6
;     end if

  casecounter = casecounter + 1
}


macro enddp {
    .__info.casecount = casecounter
    .__info.framesize = maxsize + .__info.commonframe.size
    .__info.codesize = $ - .__info.start
    .__info.casesize = caseaddr - caselist
  end if
  restore maxsize, casecounter, caselist, caseaddr
  purge err@endp
}






macro __InitInitializeEngine {
  local i,f,ci,cf
  ci equ 0
  cf equ 0

  macro initialize procname \{
    \common
    rept 1 x:ci+1 \\{
      ci equ x
      i\\#x equ procname
      proc procname
    \\}
  \}

  macro finalize procname \{
    \common
    rept 1 x:cf+1 \\{
      cf equ x
      f\\#x equ procname
      proc procname
    \\}
  \}


  macro InitializeAll \{
    \common
    rept ci x \\{
      \\forward
      if defined i\\#x
        call i\\#x
      end if
    \\}
  \}

  macro FinalizeAll \{
    \common
    rept cf x \\{
      \\reverse
      if defined f\\#x
        call f\\#x
      end if
    \\}
  \}
}

__InitInitializeEngine



;*****************************************
; Call macroses
;*****************************************


macro pushx [arg] {
common
  local f1, f2, cnt, sz
  cnt = 0
  sz = 0
  f1 = 1
  f2 equ FALSE
forward
  match any, arg \{
    match `arg, arg \\{ f2 equ TRUE \\}
  \}

  cnt = cnt + 1
  if arg eqtype ''
    virtual at 0
      db arg
      sz = sz + $
    end virtual
  end if

common
  local lbl

  if (sz > 4) | (cnt>1)
    f1 = 0
  end if

  if f1
    pushd arg
  else
    pushd lbl
  end if
  match =TRUE, f2 \{ lbl text arg \}
}




macro stdcall proc, [arg] {    ; call procedure
common
  local ..argsize
  ..argsize = 0
  if ~ arg eq
reverse
    pushx arg
    ..argsize = ..argsize + 4
common
  end if
  if defined options.CheckArguments & options.CheckArguments
    if (defined proc#.__info.argsize) & (proc#.__info.argsize <> ..argsize)
      ERROR! wrong argument count for procedure call.
    end if
  end if

  call proc
}


macro invoke proc,[arg] {   ; invoke procedure (indirect)
common
  if ~ arg eq
reverse
    pushx arg
common
  end if
  call [proc]
}


macro ccall proc,[arg]  {                ; call procedure in C calling convention
common local ..size
   ..size = 0
reverse
   pushx arg
   ..size = ..size+4
common
   call proc
   if defined proc#.__info.argsize
     if proc#.__info.argsize > 0
       add esp, proc#.__info.argsize
     end if
   else
     if ..size > 0
       add esp,..size
     end if
   end if
}


macro cinvoke proc, [arg] {               ; invoke procedure (indirect)
common local ..size
   ..size = 0
   if ~ arg eq
reverse
   pushx arg
   ..size = ..size+4
common
   end if

   call [proc]

   if defined proc#.__info.argsize
     if proc#.__info.argsize > 0
       add esp, proc#.__info.argsize
     end if
   else
     if ..size > 0
       add esp, ..size
     end if
   end if
}

;****************************************
; INT3 break point, when first = second.
;****************************************

macro BreakEq first, second {
  local .lbl
  push  eax
  mov   eax, first
  cmp   eax, second
  pop   eax
  jne   .lbl

  int3

.lbl:
}
