proc CreateFormEditor, .ptrOpenFile, .hTemplate
begin
        push    esi edi ebx

        mov     esi, [CurrentSettings]

        xor     ebx, ebx
        mov     ecx, WS_CHILD or WS_HSCROLL or WS_VSCROLL or ES_NOHIDESEL or ES_WANTRETURN
        or      ecx, [esi+TEditorSettings.AsmEditStyle]

        invoke  CreateWindowExA, WS_EX_CLIENTEDGE or WS_EX_ACCEPTFILES, cAsmEditClassName, ebx,       \
                ecx, ebx, ebx, ebx, ebx,                                                             \
                [hEditorsHost], ebx, [hInstance], ebx
        test    eax, eax
        jz      .errorcreate

        mov     edi, eax

        mov     eax, [.ptrOpenFile]
        mov     [eax+TOpenFile.hEditor], edi
        invoke  SetPropA, edi, [propOpenFileStructure], eax

; adjust the look and feel for the editor window.
        invoke  CreateFontA, [esi+TEditorSettings.FontSize], 0, 0, 0, FW_NORMAL,                 \
                            0, 0, 0, [esi+TEditorSettings.FontCharset],                         \
                            OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,           \
                            FIXED_PITCH or FF_DONTCARE, [esi+TEditorSettings.ptrFontName]
        invoke  SendMessageA, edi, WM_SETFONT, eax, ebx
        invoke  SendMessageA, edi, AEM_SETTHEME, ebx, [esi+TEditorSettings.ptrTheme]
        invoke  SendMessageA, edi, AEM_SETSYNTAXHIGHLIGHT, ebx, fasm_syntax
        invoke  SendMessageA, edi, AEM_SETPOPUPMENU, ebx, [hEditorMenu]
        invoke  SendMessageA, edi, AEM_SETBOOKMARKICON, [hBookmarkIcon], ebx

        stdcall SubclassWindow, edi, FormEditorProc

        invoke  SendMessageA, edi, CEM_LOADFILE, [.hTemplate], ebx
        test    eax, eax
        jnz     .errorload

        invoke  SendMessageA, [hEditorsHost], EHM_DOCKWINDOW, edi, dfClient

        xor     eax, eax
        pop     ebx edi esi
        return

.errorcreate:
        invoke  GetLastError
        pop     ebx edi esi
        return

.errorload:
        push    eax
        invoke  DestroyWindow, edi
        mov     [esi+TOpenFile.hEditor], ebx
        pop     eax
        pop     ebx edi esi
        return
endp


iglobal
; properties for internal use...
  property propOpenFileStructure, 'OpenFileStructure' ; This is pointer to TOpenFile structure of the edited file.
  property propWinTemplate, 'WinTemplate'
  property propVisualMode, 'VisualMode'  ; 0 - view source, 1 - view form, 2 - pure text mode.
endg


winproc FormEditorProc
begin

; prolog executed before message dispatcher
        invoke  GetPropA, [.hwnd], [propOpenFileStructure]
        mov     edi, eax

; message dispatcher
ondefault
        stc
        return

; Message handlers
;-----------------------------------------------------------------------
; Message handlers inherited from Source editor
onmessage WM_DROPFILES
onmessage WM_KEYDOWN
onmessage WM_LBUTTONDOWN
onmessage WM_RBUTTONDOWN
onmessage WM_KILLFOCUS

onmessage CCM_UPDATECCPOS
onmessage CCM_UPDATEPAPOS

onmessage WM_MEASUREITEM
onmessage WM_DRAWITEM

onmessage WM_VSCROLL
onmessage WM_HSCROLL
        stdcall SourceEditProc, [.hwnd], [.wmsg], [.wparam], [.lparam]
        return


onmessage FEM_SWITCHMODE
        invoke  GetPropA, [.hwnd], [propVisualMode]
        mov     ebx, eax
        xor     ebx, 1

; show source
        invoke  SetPropA, [.hwnd], [propVisualMode], ebx

        test    ebx, ebx
        jz      .showsource

        invoke  GetPropA, [.hwnd], [propWinTemplate]
        invoke  SetWindowPos, [eax+TWinTemplateDT.hwnd], HWND_TOP, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE or SWP_SHOWWINDOW
        clc
        return

.showsource:
        invoke  SetFocus, [.hwnd]
        clc
        return

;-----------------------------------------------------------------------
onmessage WM_SETFOCUS
        invoke  SetPropA, [.hwnd], [propVisualMode], 0
        stc
        return

;-----------------------------------------------------------------------
onmessage WM_COMMAND
        invoke  GetParent, [.hwnd]
        invoke  SendMessageA, eax, [.wmsg], [.wparam], [.lparam]
        clc
        return

;-----------------------------------------------------------------------
onmessage CEM_LOADFILE
locals
  .formname dd ?
  .forfree  dd ?
endl
        mov     [.formname], 0  ; the name is from the loaded source.

        cmp     [.wparam], 0
        jne     .loadfile

        cmp     [edi+TOpenFile.fNeverSaved], 0
        je      .loadfile

        mov     [.formname], cFormNameNone

        stdcall ConvertPath, cBaseTemplateFile
        push    eax
        stdcall LoadBinaryFile, eax
        stdcall StrDel ; from the stack
        jmp     .formloaded

cBaseTemplateFile text '%fresh%/IDE/templates/complex.asm'

; load from file
.loadfile:
        mov     eax, [.wparam]
        test    eax, eax
        jnz     @f
        mov     eax, [edi+TOpenFile.hFileName]
@@:
        stdcall StrPtr, eax
        stdcall LoadBinaryFile, eax

.formloaded:
        jnc    .loadok

; error code is in eax already
        clc
        return

.loadok:
        mov     [.forfree], eax
        mov     esi, eax

        cmp     dword [esi], ';<ff'
        jne     .errorformat

        stdcall SkipToNextLine

; Creates TWinTemplateDT for the form

        stdcall CompileSourceTemplate, esi
        mov     esi, edx
        mov     ebx, eax
        test    ebx, ebx
        jnz     .compiled

        mov     esi, [.forfree]
        jmp     .errorformat

.compiled:
        mov     [ebx+TWinTemplateDT.ptrFile], edi
        invoke  SetPropA, [.hwnd], [propWinTemplate], ebx
        stdcall CreateDesignWindow, ebx

        invoke  SendMessageA, [.hwnd], AEM_SETMODIFIED, FALSE, 0

        cmp     dword [esi], ';ff>'
        jne     .errorformat

        stdcall SkipToNextLine

.errorformat:
        invoke  SendMessageA, [.hwnd], WM_SETTEXT, 0, esi

        invoke  SendDlgItemMessageA, [hPropEditor], 101, PEM_SETCONTROL, ebx, 0
        invoke  SendDlgItemMessageA, [hPropEditor], 101, PEM_SETUPDATEPROC, 0, CreateDesignWindow
        invoke  PostMessageA, [.hwnd], FEM_SWITCHMODE, 0, 0

        cmp     [edi+TOpenFile.fNeverSaved], 0
        jne     .MRUOK
        stdcall AddToMRUList, [ptrMRUFiles], [edi+TOpenFile.hFileName]
.MRUOK:

        stdcall FreeMem, [.forfree]

        xor     eax, eax
        clc
        return



;-----------------------------------------------------------------------
onmessage _CEM_GETTEXT
        mov     ebx, [.lparam]
        test    ebx, ebx
        jz      .errortext

        mov     [ebx+TInMemoryFile.fAscii], TRUE

        invoke  GetPropA, [.hwnd], [propWinTemplate]
        test    eax, eax
        jz      .errortext

        stdcall CreateSourceTemplate, eax
        mov     esi, eax                ; string with source header.

        invoke  SendMessageA, [.hwnd], WM_GETTEXTLENGTH, 0, 0
        mov     edx, eax

        stdcall StrLen, esi
        add     edx, eax
        push    eax
        add     edx, 12

        mov     [ebx+TInMemoryFile.fileSize], edx

        add     edx, 32 ; some reserve...
        stdcall GetMem, edx
        mov     [ebx+TInMemoryFile.ptrBuffer], eax

        pop     ecx             ; length of the string
        mov     edi, eax
        mov     eax, [ebx+TInMemoryFile.fileSize]
        sub     eax, ecx
        push    eax             ; length of buffer remaining
        push    esi             ; handle of the string

        mov     eax, ';<ff'
        stosd
        mov     ax, $0a0d
        stosw

        stdcall StrPtr, esi
        mov     esi, eax

        rep movsb

        mov     eax, ';ff>'
        stosd
        mov     ax, $0a0d
        stosw

        pop     esi
        pop     ecx
        add     ecx, 2  ; reserve for terminating NULL

        invoke  SendMessageA, [.hwnd], WM_GETTEXT, ecx, edi

        stdcall StrDel, esi     ; delete temporary string.

        xor     eax, eax
        clc
        return

.errortext:
        xor     eax, eax
        inc     eax
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_CHANGENOTIFY
        invoke  GetParent, [.hwnd]
        invoke  SendMessageA, eax, EHM_FILECHANGED, edi, 0
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_FOCUSFILE

        invoke  GetParent, [.hwnd]
        invoke  SendMessageA, eax, EHM_FOCUSWINDOW, [.hwnd], 0
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_CANACTION
  locals
    .aepos AEPOS
  endl
        mov     eax, [.lparam]
        cmp     eax, WM_CUT
        je      .checkcopycut
        cmp     eax, WM_COPY
        je      .checkcopycut
        cmp     eax, WM_CLEAR
        je      .checkcopycut
        cmp     eax, WM_UNDO
        je      .checkundo
        cmp     eax, EM_UNDO
        je      .checkundo
        cmp     eax, WM_PASTE
        je      .checkpaste
        xor     eax, eax
        clc
        return

.checkcopycut:

        invoke  GetPropA, [.hwnd], [propVisualMode]
        test    eax, eax
        jz      .checksource

        invoke  GetPropA, [.hwnd], [propWinTemplate]
        stdcall GetSelectionList, eax
        test    eax, eax
        jz      .exitcopycheck

        mov     eax, [eax+TArray.count]

.exitcopycheck:
        clc
        return

.checksource:
        lea     esi, [.aepos]
        invoke  SendMessageA, [.hwnd], AEM_GETPOS, esi, 0
        mov     eax, [esi+AEPOS.caretPosition]
        mov     ecx, [esi+AEPOS.caretLine]
        cmp     [esi+AEPOS.selectionPosition], eax
        jne     .can_copy
        cmp     [esi+AEPOS.selectionLine], ecx
        jne     .can_copy

        xor     eax, eax
        clc
        return

.can_copy:
        xor     eax, eax
        inc     eax
        clc
        return

.checkundo:
        invoke  SendMessageA, [.hwnd], EM_CANUNDO, 0, 0
        clc
        return

.checkpaste:
        invoke  IsClipboardFormatAvailable, CF_TEXT
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_EDITACTION

        invoke  GetPropA, [.hwnd], [propVisualMode]
        test    eax, eax
        jnz     .formmode

.sourcemode:
        invoke  SendMessageA, [.hwnd], [.lparam], 0, 0
        clc
        return

;-------------------------------------
; Common actions for form edit mode
;-------------------------------------
.formmode:
        cmp     [.lparam], WM_COPY
        je      .copycontrol

        cmp     [.lparam], WM_CUT
        je      .cutcontrol

        cmp     [.lparam], WM_CLEAR
        je      .delcontrol

        cmp     [.lparam], WM_PASTE
        je      .pastecontrol

        clc
        return

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.pastecontrol:
  locals
    .hmem dd ?
    .ptrTarget dd ?
  endl
        invoke  IsClipboardFormatAvailable, CF_TEXT
        test    eax, eax
        jz      .endpaste

        invoke  OpenClipboard, NULL

        invoke  GetClipboardData, CF_TEXT
        test    eax, eax
        jz      .closepastecb

        mov     [.hmem], eax
        invoke  GlobalLock, eax
        mov     esi, eax

        invoke  GetPropA, [.hwnd], [propWinTemplate]
        mov     edi, eax

; search for paste target
        stdcall GetSelectionList, edi
        mov     ecx, [eax+TArray.count]
        jecxz   .compileloop

        mov     edi, [eax+TArray.array]

; Compile windows from clipboard...
.compileloop:
        stdcall CompileSourceTemplate, esi
        test    eax, eax
        jz      .endcompileloop

        mov     esi, edx
        mov     ebx, eax

        lea     ecx, [edi+TWinTemplateDT.children]
        stdcall AddArrayItems, [edi+TWinTemplateDT.children], 1
        mov     [edi+TWinTemplateDT.children], edx
        mov     [eax], ebx
        mov     [ebx+TWinTemplateDT.parent], edi

; Check for coordinates to be inside client area of the target
        mov     cx, [ebx+TWinTemplateDT.left]
        mov     dx, [ebx+TWinTemplateDT.top]

.adjustx:
        cmp     cx, [edi+TWinTemplateDT.width]
        jb      .adjusty
        sub     cx, [edi+TWinTemplateDT.width]
        jmp     .adjustx

.adjusty:
        cmp     dx, [edi+TWinTemplateDT.height]
        jb      .xyok
        sub     dx, [edi+TWinTemplateDT.height]
        jmp     .adjusty

; Creates windows instances...
.xyok:
        mov     [ebx+TWinTemplateDT.left], cx
        mov     [ebx+TWinTemplateDT.top], dx
        stdcall CreateDesignWindow, ebx
        invoke  SendMessageA, [.hwnd], AEM_SETMODIFIED, TRUE, 0
        jmp     .compileloop

.endcompileloop:
        invoke  GlobalUnlock, [.hmem]

.closepastecb:
        invoke  CloseClipboard

.endpaste:
        clc
        return

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.cutcontrol:
        invoke  SendMessageA, [.hwnd], _CEM_EDITACTION, 0, WM_COPY

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.delcontrol:
        invoke  GetPropA, [.hwnd], [propWinTemplate]
        push    eax
        stdcall GetSelectionList, eax

        mov     ebx, [eax+TArray.count]
        lea     esi, [eax+TArray.array]

.delloop:
        dec     ebx
        js      .enddel

        stdcall DestroyDesignWindow, [esi+4*ebx], TRUE
        jmp     .delloop

.enddel:
        pop     eax
        stdcall GetSizer
        invoke  SendMessageA, eax, SZM_UNSELECTALL, 0, 0
        invoke  SendMessageA, [.hwnd], AEM_SETMODIFIED, TRUE, 0
        clc
        return

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
.copycontrol:
        stdcall StrNew
        mov     ebx, eax

        invoke  GetPropA, [.hwnd], [propWinTemplate]
        stdcall GetSelectionList, eax

        mov     edi, [eax+TArray.count]
        lea     esi, [eax+TArray.array]

.copyselected:
        dec     edi
        js      .endcopyloop

        stdcall CreateSourceTemplate, [esi]
        stdcall StrCat, ebx, eax
        stdcall StrDel, eax

        add     esi, 4
        jmp     .copyselected

.endcopyloop:
        stdcall StrLen, ebx
        test    eax, eax
        jz      .endcopy

        lea     esi, [eax+1]

        invoke  OpenClipboard, NULL
        test    eax, eax
        jz      .endcopy

        invoke  EmptyClipboard

        invoke  GlobalAlloc, GHND or GMEM_DDESHARE, esi
        test    eax, eax
        jz      .closecb

        mov     [.hmem], eax

        invoke  GlobalLock, eax
        mov     edi, eax

        mov     ecx, esi
        stdcall StrPtr, ebx
        mov     esi, eax
        rep movsb
        invoke  GlobalUnlock, [.hmem]

        invoke  SetClipboardData, CF_TEXT, [.hmem]

.closecb:
        invoke  CloseClipboard

.endcopy:
        stdcall StrDel, ebx
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_GETMODIFIED
        invoke   SendMessageA, [.hwnd], AEM_GETMODIFIED, 0, 0
        clc
        return

;-----------------------------------------------------------------------
onmessage _CEM_SETMODIFIED
        invoke  SendMessageA, [.hwnd], AEM_SETMODIFIED, [.wparam], [.lparam]
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_GETFILE
        mov     eax, edi
        clc
        return


;-----------------------------------------------------------------------
onmessage _CEM_GETASMEDIT
        mov     eax, [.hwnd]
        clc
        return

;-----------------------------------------------------------------------
onmessage WM_DESTROY
        invoke  GetPropA, [.hwnd], [propWinTemplate]
        stdcall DestroyDesignWindow, eax, TRUE
        stdcall SourceEditProc, [.hwnd], [.wmsg], [.wparam], [.lparam]
        return
endwp




;------------------------------------------------------------------------
; Several callback procedures for EnumSelections procedure "formsDT.asm"
;------------------------------------------------------------------------
proc CopySelectedVisual, .dummy, .ptrList, .hString
begin
        push    esi edi ebx

        mov     esi, [.ptrList]
        mov     ebx, [esi+TList.Count]
        mov     esi, [esi+TList.ptrItems]

.loop:
        dec     ebx
        js      .endloop

        stdcall CreateSourceTemplate, [esi]
        stdcall StrCat, [.hString], eax
        stdcall StrDel, eax

        add     esi, 4
        jmp     .loop

.endloop:
        pop     ebx edi esi
        return
endp



proc UnselectAll, .ptrWinTemplateDT
begin
        mov     eax, [.ptrWinTemplateDT]
        stdcall GetSizer
        invoke  SendMessageA, eax, SZM_UNSELECTALL, 0, 0
        return
endp